/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.model.framestore;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.util.*;

public class JournalingFrameStoreHandler extends AbstractFrameStoreInvocationHandler {
    private PrintWriter _writer = new PrintWriter(System.out);
    private boolean recordQueries = false;
    private String userName = SystemUtilities.getUserName();
    private int transactionLevel;

    private boolean inTransaction() {
        return transactionLevel != 0;
    }

    public void setWriter(PrintWriter writer) {
        _writer = writer;
    }

    public void start(String userName, URI uri) {
        this.userName = userName;
        File file = new File(uri);
        _writer = FileUtilities.createPrintWriter(file, true);
    }

    public void stop() {
        _writer = null;
    }

    public void handleClose() {
        _writer.flush();
        _writer = null;
    }

    private void handleTransactions(Method method) {
        String name = method.getName();
        if (contains(name, "beginTransaction")) {
            ++transactionLevel;
            Log.trace("begin transaction:" + transactionLevel, this, "handleTransactions", method);
        } else if (contains(name, "Transaction")) {
            --transactionLevel;
            Log.trace("end transaction: " + transactionLevel, this, "handleTransactions", method);
        }
    }

    private static boolean contains(String s, String substring) {
        return s.indexOf(substring) != -1;
    }

    public Object handleInvoke(Method method, Object[] args) {
        boolean record = doRecord(method);
        handleTransactions(method);
        if (record) {
            recordCall(method, args);
        }

        Object o = invoke(method, args);
        if (record) {
            recordReturn(method, o);
        }
        return o;
    }

    private String getFrameName(Frame frame) {
        return getDelegate().getFrameName(frame);
    }

    protected boolean doRecord(Method method) {
        return !inTransaction() && (recordQueries || isModification(method));
    }

    protected void recordCall(Method method, Object[] args) {
        StringBuffer text = new StringBuffer();
        text.append(new Date());
        text.append(", ");
        text.append(userName);
        text.append(", ");
        text.append(method.getName());
        text.append("(");
        if (args != null) {
            for (int i = 0; i < args.length; ++i) {
                if (i != 0) {
                    text.append(", ");
                }
                text.append(argumentToString(args[i]));
            }
        }
        text.append(")");
        _writer.print(text);
    }

    protected void recordReturn(Method method, Object returnValue) {
        StringBuffer text = new StringBuffer();
        if (method.getReturnType() != Void.TYPE) {
            text.append(" return ");
            text.append(argumentToString(returnValue));
        }
        _writer.println(text);
        _writer.flush();
    }

    protected String argumentToString(Object o) {
        String s;
        if (o == null) {
            s = "null";
        } else if (o instanceof Collection) {
            s = collectionToString((Collection) o);
        } else {
            s = toString(o);
        }
        return s;
    }

    protected String toString(Object o) {
        String s;
        if (o instanceof Frame) {
            String clas;
            if (o instanceof Cls) {
                clas = "Class";
            } else if (o instanceof Slot) {
                clas = "Slot";
            } else if (o instanceof Facet) {
                clas = "Facet";
            } else {
                clas = "SimpleInstance";
            }
            s = clas + "(" + getFrameName((Frame) o) + ")";
        } else if (o instanceof String) {
            s = "\"" + o.toString() + "\"";
        } else {
            s = o.toString();
        }
        return s;
    }

    protected String collectionToString(Collection c) {
        boolean first = true;
        StringBuffer text = new StringBuffer("{");
        Iterator i = c.iterator();
        while (i.hasNext()) {
            if (first) {
                first = false;
            } else {
                text.append(", ");
            }
            Object o = i.next();
            text.append(toString(o));
        }
        text.append("}");
        return text.toString();
    }
}

/*
public class TracingFrameStore extends DelegatingFrameStore {
    private static final Object VOID = new Object();

    public Cls createCls(String name, Collection directTypes, Collection directSuperclasses, Session session) throws Exception {
        traceCall("createCls", name, directTypes, directSuperclasses, session);
        Cls cls = getDelegate().createCls(name, directTypes, directSuperclasses, session);
        traceReturn(cls);
        return cls;
    }
    public void deleteFrame(Frame frame, Session session) throws Exception {
        traceCall("deleteFrame", frame, session);
        getDelegate().deleteFrame(frame, session);
        traceReturn();
    }
    public SimpleInstance createSimpleInstance(String name, Collection directTypes, Session session) throws Exception {
        traceCall("createSimpleInstance", name, directTypes, session);
        SimpleInstance simpleInstance = getDelegate().createSimpleInstance(name, directTypes, session);
        traceReturn(simpleInstance);
        return simpleInstance;
    }
    public Slot createSlot(String name, Collection directTypes, Collection directSuperslots, Session session) throws Exception {
        traceCall("createSlot", name, directTypes, directSuperslots, session);
        Slot slot = getDelegate().createSlot(name, directTypes, directSuperslots, session);
        traceReturn(slot);
        return slot;
    }
    public Facet createFacet(String name, Collection directTypes, Session session) throws Exception {
        traceCall("createFacet", name, directTypes, session);
        Facet facet = getDelegate().createFacet(name, directTypes, session);
        traceReturn(facet);
        return facet;
    }
    public void setDirectOwnSlotValues(Frame frame, Slot slot, Collection values, Session session) throws Exception {
        traceCall("setOwnSlotValues", frame, slot, values, session);
        getDelegate().setDirectOwnSlotValues(frame, slot, values, session);
        traceReturn();
    }
    public void setDirectTemplateSlotValues(Cls cls, Slot slot, Collection values, Session session) throws Exception {
        traceCall("setTemplateSlotValues", cls, slot, values, session);
        getDelegate().setDirectTemplateSlotValues(cls, slot, values, session);
        traceReturn();
    }
    public void setDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet, Collection values, Session session) throws Exception {
        traceCall("setTemplateFacetValues", cls, slot, facet, values, session);
        getDelegate().setDirectTemplateFacetValues(cls, slot, facet, values, session);
        traceReturn();
    }
    public void addDirectTemplateSlot(Cls cls, Slot slot, Session session) throws Exception {
        traceCall("addDirectTemplateSlot", cls, slot, session);
        getDelegate().addDirectTemplateSlot(cls, slot, session);
        traceReturn();
    }
    public void removeDirectTemplateSlot(Cls cls, Slot slot, Session session) throws Exception {
        traceCall("removeDirectTemplateSlot", cls, slot, session);
        getDelegate().removeDirectTemplateSlot(cls, slot, session);
        traceReturn();
    }
    public void moveDirectTemplateSlot(Cls cls, Slot slot, int index, Session session) throws Exception {
        traceCall("moveDirectTemplateSlot", cls, slot, index, session);
        getDelegate().moveDirectTemplateSlot(cls, slot, index, session);
        traceReturn();
    }
    public void addDirectSuperclass(Cls cls, Cls superclass, Session session) throws Exception {
        traceCall("addDirectSuperclass", cls, superclass, session);
        getDelegate().addDirectSuperclass(cls, superclass, session);
        traceReturn();
    }
    public void removeDirectSuperclass(Cls cls, Cls superclass, Session session) throws Exception {
        traceCall("removeDirectSuperclass", cls, superclass, session);
        getDelegate().removeDirectSuperclass(cls, superclass, session);
        traceReturn();
    }
    public void moveDirectSubclass(Cls cls, Cls subclass, int index, Session session) throws Exception {
        traceCall("moveDirectSubclass", cls, subclass, index, session);
        getDelegate().moveDirectSubclass(cls, subclass, index, session);
        traceReturn();
    }
    public void addDirectSuperslot(Slot slot, Slot superslot, Session session) throws Exception {
        traceCall("addDirectSuperslot", slot, superslot, session);
        getDelegate().addDirectSuperslot(slot, superslot, session);
        traceReturn();
    }
    public void removeDirectSuperslot(Slot slot, Slot superslot, Session session) throws Exception {
        traceCall("removeDirectSuperslot", slot, superslot, session);
        getDelegate().removeDirectSuperslot(slot, superslot, session);
        traceReturn();
    }
    public void addDirectType(Instance instance, Cls type, Session session) throws Exception {
        traceCall("addDirectType", instance, type, session);
        getDelegate().addDirectType(instance, type, session);
        traceReturn();
    }
    public void removeDirectType(Instance instance, Cls type, Session session) throws Exception {
        traceCall("removeDirectType", instance, type, session);
        getDelegate().removeDirectType(instance, type, session);
        traceReturn();
    }

    private void traceCall(String method, Session session) {
        traceCall(method, new Object[]{}, session);
    }
    private void traceCall(String method, Object arg1, Session session) {
        traceCall(method, new Object[]{arg1}, session);
    }
    private void traceCall(String method, Object arg1, Object arg2, Session session) {
        traceCall(method, new Object[]{arg1, arg2}, session);
    }
    private void traceCall(String method, Object arg1, Object arg2, Object arg3, Session session) {
        traceCall(method, new Object[]{arg1, arg2, arg3}, session);
    }
    private void traceCall(String method, Object arg1, Object arg2, int arg3, Session session) {
        traceCall(method, new Object[]{arg1, arg2, new Integer(arg3)}, session);
    }
    private void traceCall(String method, Object arg1, Object arg2, Object arg3, Object arg4, Session session) {
        traceCall(method, new Object[]{arg1, arg2, arg3, arg4}, session);
    }
    private void traceCall(String method, Object[] args, Session session) {
        StringBuffer text = new StringBuffer(method);
        text.append("(");
        for (int i = 0; i < args.length; ++i) {
            if (i != 0) {
                text.append(", ");
            }
            text.append(argumentToString(args[i]));
        }
        text.append(", ");
        text.append(session);
        text.append(")");
        System.out.print(text);
    }

    private void traceReturn() {
        traceReturn(VOID);
    }

    private void traceReturn(Object returnValue) {
        StringBuffer text = new StringBuffer();
        if (returnValue != VOID) {
            text.append(" return ");
            text.append(argumentToString(returnValue));
        }
        System.out.println(text);
    }

    private static String argumentToString(Object o) {
        String s;
        if (o == null) {
            s = "null";
        } else if (o instanceof Collection) {
            s = collectionToString((Collection) o);
        } else {
            s = o.toString();
        }
        return s;
    }

    private static String collectionToString(Collection c) {
        boolean first = true;
        StringBuffer text = new StringBuffer("{");
        Iterator i = c.iterator();
        while (i.hasNext()) {
            if (first) {
                first = false;
            } else {
                text.append(",");
            }
            Object o = i.next();
            text.append(o);
        }
        text.append("}");
        return text.toString();
    }

    public Set getClses(Session session) throws Exception {
        traceCall("getClses", session);
        Set result = getDelegate().getClses(session);
        traceReturn(result);
        return result;
    }
    public Set getSlots(Session session) throws Exception {
        traceCall("getSlots", session);
        Set result = getDelegate().getSlots(session);
        traceReturn(result);
        return result;
    }
    public Set getFacets(Session session) throws Exception {
        traceCall("getFacets", session);
        Set result = getDelegate().getFacets(session);
        traceReturn(result);
        return result;
    }
    public Frame getFrame(String name, Session session) throws Exception {
        traceCall("getFrame", name, session);
        Frame result = getDelegate().getFrame(name, session);
        traceReturn(result);
        return result;
    }
    public Set getFrames(Session session) throws Exception {
        traceCall("getFrames", session);
        Set result = getDelegate().getFrames(session);
        traceReturn(result);
        return result;
    }
    public Set getOwnSlots(Frame frame, Session session) throws Exception {
        traceCall("getOwnSlots", frame, session);
        Set result = getDelegate().getOwnSlots(frame, session);
        traceReturn(result);
        return result;
    }
    public Collection getOwnSlotValues(Frame frame, Slot slot, Session session) throws Exception {
        traceCall("getOwnSlotValues", frame, slot, session);
        Collection result = getDelegate().getOwnSlotValues(frame, slot, session);
        traceReturn(result);
        return result;
    }
    public List getDirectOwnSlotValues(Frame frame, Slot slot, Session session) throws Exception {
        traceCall("getDirectOwnSlotValues", frame, slot, session);
        List result = getDelegate().getDirectOwnSlotValues(frame, slot, session);
        traceReturn(result);
        return result;
    }
    public Set getTemplateSlots(Cls cls, Session session) throws Exception {
        traceCall("getTemplateSlots", cls, session);
        Set result = getDelegate().getTemplateSlots(cls, session);
        traceReturn(result);
        return result;
    }
    public List getDirectTemplateSlots(Cls cls, Session session) throws Exception {
        traceCall("getDirectTemplateSlots", cls, session);
        List result = getDelegate().getDirectTemplateSlots(cls, session);
        traceReturn(result);
        return result;
    }
    public Set getTemplateSlotValues(Cls cls, Slot slot, Session session) throws Exception {
        traceCall("getTemplateSlotValues", cls, slot, session);
        Set result = getDelegate().getTemplateSlotValues(cls, slot, session);
        traceReturn(result);
        return result;
    }
    public List getDirectTemplateSlotValues(Cls cls, Slot slot, Session session) throws Exception {
        traceCall("getDirectTemplateSlotValues", cls, slot, session);
        List result = getDelegate().getDirectTemplateSlotValues(cls, slot, session);
        traceReturn(result);
        return result;
    }
    public Set getTemplateFacets(Cls cls, Slot slot, Session session) throws Exception {
        traceCall("getTemplateFacets", cls, slot, session);
        Set result = getDelegate().getTemplateFacets(cls, slot, session);
        traceReturn(result);
        return result;
    }
    public Set getTemplateFacetValues(Cls cls, Slot slot, Facet facet, Session session) throws Exception {
        traceCall("getTemplateFacetValues", cls, slot, facet, session);
        Set result = getDelegate().getTemplateFacetValues(cls, slot, facet, session);
        traceReturn(result);
        return result;
    }
    public List getDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet, Session session) throws Exception {
        traceCall("getDirectTemplateFacetValues", cls, slot, facet, session);
        List result = getDelegate().getDirectTemplateFacetValues(cls, slot, facet, session);
        traceReturn(result);
        return result;
    }
    public List getDirectSuperclasses(Cls cls, Session session) throws Exception {
        traceCall("getDirectSuperclasses", cls, session);
        List result = getDelegate().getDirectSuperclasses(cls, session);
        traceReturn(result);
        return result;
    }
    public Set getSuperclasses(Cls cls, Session session) throws Exception {
        traceCall("getSuperclasses", cls, session);
        Set result = getDelegate().getSuperclasses(cls, session);
        traceReturn(result);
        return result;
    }
    public List getDirectSubclasses(Cls cls, Session session) throws Exception {
        traceCall("getDirectSubclasses", cls, session);
        List result = getDelegate().getDirectSubclasses(cls, session);
        traceReturn(result);
        return result;
    }
    public Set getSubclasses(Cls cls, Session session) throws Exception {
        traceCall("getSubclasses", cls, session);
        Set result = getDelegate().getSubclasses(cls, session);
        traceReturn(result);
        return result;
    }
    public List getDirectSuperslots(Slot slot, Session session) throws Exception {
        traceCall("getDirectSuperslots", slot, session);
        List result = getDelegate().getDirectSuperslots(slot, session);
        traceReturn(result);
        return result;
    }
    public Set getSuperslots(Slot slot, Session session) throws Exception {
        traceCall("getSuperslots", slot, session);
        Set result = getDelegate().getSuperslots(slot, session);
        traceReturn(result);
        return result;
    }
    public List getDirectSubslots(Slot slot, Session session) throws Exception {
        traceCall("getDirectSubslots", slot, session);
        List result = getDelegate().getDirectSubslots(slot, session);
        traceReturn(result);
        return result;
    }
    public Set getSubslots(Slot slot, Session session) throws Exception {
        traceCall("getSubslots", slot, session);
        Set result = getDelegate().getSubslots(slot, session);
        traceReturn(result);
        return result;
    }
    public List getDirectTypes(Instance instance, Session session) throws Exception {
        traceCall("getDirectTypes", instance, session);
        List result = getDelegate().getDirectTypes(instance, session);
        traceReturn(result);
        return result;
    }
    public Set getTypes(Instance instance, Session session) throws Exception {
        traceCall("getTypes", instance, session);
        Set result = getDelegate().getTypes(instance, session);
        traceReturn(result);
        return result;
    }
    public List getDirectInstances(Cls cls, Session session) throws Exception {
        traceCall("getDirectInstances", cls, session);
        List result = getDelegate().getDirectInstances(cls, session);
        traceReturn(result);
        return result;
    }
    public Set getInstances(Cls cls, Session session) throws Exception {
        traceCall("getInstances", cls, session);
        Set result = getDelegate().getInstances(cls, session);
        traceReturn(result);
        return result;
    }
    public List getEvents(Session session) throws Exception {
        traceCall("getEvents", session);
        List result = getDelegate().getEvents(session);
        traceReturn(result);
        return result;
    }
    public Set getReferences(Object value, Session session) throws Exception {
        traceCall("getReferences", value, session);
        Set result = getDelegate().getReferences(value, session);
        traceReturn(result);
        return result;
    }
    public Set getMatchingReferences(String value, Session session) throws Exception {
        traceCall("getMatchingReferences", value, session);
        Set result = getDelegate().getMatchingReferences(value, session);
        traceReturn(result);
        return result;
    }
    public Set getFramesWithOwnSlotValue(Slot slot, Object value, Session session) throws Exception {
        traceCall("getFramesWithOwnSlotValue", slot, value, session);
        Set result = getDelegate().getFramesWithOwnSlotValue(slot, value, session);
        traceReturn(result);
        return result;
    }
    public Set getFramesWithMatchingOwnSlotValue(Slot slot, String value, Session session) throws Exception {
        traceCall("getFramesWithMatchingOwnSlotValue", slot, value, session);
        Set result = getDelegate().getFramesWithMatchingOwnSlotValue(slot, value, session);
        traceReturn(result);
        return result;
    }
    public Set executeQuery(Query query, Session session) throws Exception {
        traceCall("executeQuery", query, session);
        Set result = getDelegate().executeQuery(query, session);
        traceReturn(result);
        return result;
    }

    public String getFrameName(Frame frame, Session session) throws Exception {
        traceCall("getFrameName", frame, session);
        String name = getDelegate().getFrameName(frame, session);
        traceReturn(name);
        return name;
    }

    public void setFrameName(Frame frame, String name, Session session) throws Exception {
        traceCall("setFrameName", frame, name, session);
        getDelegate().setFrameName(frame, name, session);
        traceReturn();
    }

    public Facet getAssociatedFacet(Slot slot, Session session) throws Exception {
        traceCall("getAssociateFacet", slot, session);
        Facet facet = getDelegate().getAssociatedFacet(slot, session);
        traceReturn(facet);
        return facet;
    }

    public Slot getAssociatedSlot(Facet facet, Session session) throws Exception {
        traceCall("getAssociateSlot", facet, session);
        Slot slot = getDelegate().getAssociatedSlot(facet, session);
        traceReturn(slot);
        return slot;
    }

    public void setAssociatedFacet(Slot slot, Facet facet, Session session) throws Exception {
        traceCall("setAssociatedFacet", slot, facet, session);
        getDelegate().setAssociatedFacet(slot, facet, session);
        traceReturn();
    }

    public Slot getInverseSlot(Slot slot, Session session) throws Exception {
        traceCall("getInverseSlot", slot, session);
        Slot inverse = getDelegate().getInverseSlot(slot, session);
        traceReturn(inverse);
        return inverse;
    }

    public void setInverseSlot(Slot slota, Slot slotb, Session session) throws Exception {
        traceCall("setInverseSlot", slota, slotb, session);
        getDelegate().setInverseSlot(slota, slotb, session);
        traceReturn();
    }
}
*/
