/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.server;

import java.rmi.*;
import java.util.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.model.framestore.*;
import edu.stanford.smi.protege.model.query.*;
import edu.stanford.smi.protege.util.*;

public class RemoteClientFrameStore implements FrameStore {
    private KnowledgeBase kb;
    private RemoteSession session;
    private RemoteServerFrameStore delegate;

    public RemoteClientFrameStore(String host, String user, String password, String projectName, KnowledgeBase kb) {
        try {
            RemoteServer server = (RemoteServer) Naming.lookup("//" + host + "/" + Server.getBoundName());
            String machine = SystemUtilities.getMachineName();
            RemoteSession session = server.openSession(user, machine, password);
            RemoteServerProject project = server.openProject(projectName, session);
            RemoteServerFrameStore store = project.getDomainKbFrameStore(session);
            init(store, session, kb);
        } catch (Exception e) {
            Log.exception(e, this, "RemoteClientFrameStoreHandler", host, user, password, projectName);
        }
    }

    public RemoteClientFrameStore(RemoteServerFrameStore delegate, RemoteSession session, KnowledgeBase kb) {
        init(delegate, session, kb);
    }

    private void init(RemoteServerFrameStore delegate, RemoteSession session, KnowledgeBase kb) {
        this.session = session;
        this.kb = kb;
        this.delegate = delegate;
    }

    public void setDelegate(FrameStore delegate) {
        throw new UnsupportedOperationException();
    }

    public FrameStore getDelegate() {
        return null;
    }

    public void reinitialize() {
        // do nothing
    }

    private RuntimeException convertException(Exception e) {
        return new RuntimeException(e);
    }

    public int getClsCount() {
        try {
            return delegate.getClsCount(session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public int getSlotCount() {
        try {
            return delegate.getSlotCount(session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public int getFacetCount() {
        try {
            return delegate.getFacetCount(session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public int getSimpleInstanceCount() {
        try {
            return delegate.getSimpleInstanceCount(session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public int getFrameCount() {
        try {
            return delegate.getFrameCount(session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    private void localize(Frame frame) {
        if (frame != null) {
            LocalizeUtils.localizeFrame(frame, kb);
        }
    }

    private void localize(Collection collection) {
        LocalizeUtils.localizeCollection(collection, kb);
    }

    public Set getClses() {
        try {
            Set clses = delegate.getClses(session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getSlots() {
        try {
            Set slots = delegate.getSlots(session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getFacets() {
        try {
            Set facets = delegate.getFacets(session);
            localize(facets);
            return facets;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getFrames() {
        try {
            Set frames = delegate.getFrames(session);
            localize(frames);
            return frames;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Frame getFrame(FrameID id) {
        try {
            Frame frame = delegate.getFrame(id, session);
            localize(frame);
            return frame;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Frame getFrame(String name) {
        try {
            Frame frame = delegate.getFrame(name, session);
            localize(frame);
            return frame;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public String getFrameName(Frame frame) {
        try {
            return delegate.getFrameName(frame, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void setFrameName(Frame frame, String name) {
        try {
            delegate.setFrameName(frame, name, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Cls createCls(
        FrameID id,
        String name,
        Collection directTypes,
        Collection directSuperclasses,
        boolean loadDefaultValues) {
        try {
            Cls cls = delegate.createCls(id, name, directTypes, directSuperclasses, loadDefaultValues, session);
            localize(cls);
            return cls;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Slot createSlot(
        FrameID id,
        String name,
        Collection directTypes,
        Collection directSuperslots,
        boolean loadDefaultValues) {
        try {
            Slot slot = delegate.createSlot(id, name, directTypes, directSuperslots, loadDefaultValues, session);
            localize(slot);
            return slot;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Facet createFacet(FrameID id, String name, Collection directTypes, boolean loadDefaultValues) {
        try {
            Facet facet = delegate.createFacet(id, name, directTypes, loadDefaultValues, session);
            localize(facet);
            return facet;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public SimpleInstance createSimpleInstance(
        FrameID id,
        String name,
        Collection directTypes,
        boolean loadDefaultValues) {
        try {
            SimpleInstance simpleInstance =
                delegate.createSimpleInstance(id, name, directTypes, loadDefaultValues, session);
            localize(simpleInstance);
            return simpleInstance;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void deleteCls(Cls cls) {
        try {
            delegate.deleteCls(cls, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void deleteSlot(Slot slot) {
        try {
            delegate.deleteSlot(slot, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void deleteFacet(Facet facet) {
        try {
            delegate.deleteFacet(facet, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void deleteSimpleInstance(SimpleInstance simpleInstance) {
        try {
            delegate.deleteSimpleInstance(simpleInstance, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getOwnSlots(Frame frame) {
        try {
            Set slots = delegate.getOwnSlots(frame, session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Collection getOwnSlotValues(Frame frame, Slot slot) {
        try {
            Collection values = delegate.getOwnSlotValues(frame, slot, session);
            localize(values);
            return values;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectOwnSlotValues(Frame frame, Slot slot) {
        try {
            List values = delegate.getDirectOwnSlotValues(frame, slot, session);
            localize(values);
            return values;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public int getDirectOwnSlotValuesCount(Frame frame, Slot slot) {
        try {
            return delegate.getDirectOwnSlotValuesCount(frame, slot, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void setDirectOwnSlotValues(Frame frame, Slot slot, Collection values) {
        try {
            delegate.setDirectOwnSlotValues(frame, slot, values, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getOwnFacets(Frame frame, Slot slot) {
        try {
            Set facets = delegate.getOwnFacets(frame, slot, session);
            localize(facets);
            return facets;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Collection getOwnFacetValues(Frame frame, Slot slot, Facet facet) {
        try {
            Collection values = delegate.getOwnFacetValues(frame, slot, facet, session);
            localize(values);
            return values;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getTemplateSlots(Cls cls) {
        try {
            Set slots = delegate.getTemplateSlots(cls, session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectTemplateSlots(Cls cls) {
        try {
            List slots = delegate.getDirectTemplateSlots(cls, session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectDomain(Slot slot) {
        try {
            List domain = delegate.getDirectDomain(slot, session);
            localize(domain);
            return domain;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getDomain(Slot slot) {
        try {
            Set domain = delegate.getDomain(slot, session);
            localize(domain);
            return domain;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getOverriddenTemplateSlots(Cls cls) {
        try {
            Set slots = delegate.getOverriddenTemplateSlots(cls, session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getDirectlyOverriddenTemplateSlots(Cls cls) {
        try {
            Set slots = delegate.getDirectlyOverriddenTemplateSlots(cls, session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void addDirectTemplateSlot(Cls cls, Slot slot) {
        try {
            delegate.addDirectTemplateSlot(cls, slot, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void removeDirectTemplateSlot(Cls cls, Slot slot) {
        try {
            delegate.removeDirectTemplateSlot(cls, slot, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void moveDirectTemplateSlot(Cls cls, Slot slot, int index) {
        try {
            delegate.moveDirectTemplateSlot(cls, slot, index, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Collection getTemplateSlotValues(Cls cls, Slot slot) {
        try {
            Collection values = delegate.getTemplateSlotValues(cls, slot, session);
            localize(values);
            return values;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectTemplateSlotValues(Cls cls, Slot slot) {
        try {
            List values = delegate.getDirectTemplateSlotValues(cls, slot, session);
            localize(values);
            return values;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void setDirectTemplateSlotValues(Cls cls, Slot slot, Collection values) {
        try {
            delegate.setDirectTemplateSlotValues(cls, slot, values, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getTemplateFacets(Cls cls, Slot slot) {
        try {
            Set facets = delegate.getTemplateFacets(cls, slot, session);
            localize(facets);
            return facets;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getOverriddenTemplateFacets(Cls cls, Slot slot) {
        try {
            Set facets = delegate.getOverriddenTemplateFacets(cls, slot, session);
            localize(facets);
            return facets;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getDirectlyOverriddenTemplateFacets(Cls cls, Slot slot) {
        try {
            Set facets = delegate.getDirectlyOverriddenTemplateFacets(cls, slot, session);
            localize(facets);
            return facets;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void removeDirectTemplateFacetOverrides(Cls cls, Slot slot) {
        try {
            delegate.removeDirectTemplateFacetOverrides(cls, slot, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Collection getTemplateFacetValues(Cls cls, Slot slot, Facet facet) {
        try {
            Collection values = delegate.getTemplateFacetValues(cls, slot, facet, session);
            localize(values);
            return values;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet) {
        try {
            List values = delegate.getDirectTemplateFacetValues(cls, slot, facet, session);
            localize(values);
            return values;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void setDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet, Collection values) {
        try {
            delegate.setDirectTemplateFacetValues(cls, slot, facet, values, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }
    public List getDirectSuperclasses(Cls cls) {
        try {
            List clses = delegate.getDirectSuperclasses(cls, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }
    public Set getSuperclasses(Cls cls) {
        try {
            Set clses = delegate.getSuperclasses(cls, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectSubclasses(Cls cls) {
        try {
            List clses = delegate.getDirectSubclasses(cls, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getSubclasses(Cls cls) {
        try {
            Set clses = delegate.getSubclasses(cls, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void addDirectSuperclass(Cls cls, Cls superclass) {
        try {
            delegate.addDirectSuperclass(cls, superclass, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void removeDirectSuperclass(Cls cls, Cls superclass) {
        try {
            delegate.removeDirectSuperclass(cls, superclass, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void moveDirectSubclass(Cls cls, Cls subclass, int index) {
        try {
            delegate.moveDirectSubclass(cls, subclass, index, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectSuperslots(Slot slot) {
        try {
            List slots = delegate.getDirectSuperslots(slot, session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getSuperslots(Slot slot) {
        try {
            Set slots = delegate.getSuperslots(slot, session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectSubslots(Slot slot) {
        try {
            List slots = delegate.getDirectSubslots(slot, session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getSubslots(Slot slot) {
        try {
            Set slots = delegate.getSubslots(slot, session);
            localize(slots);
            return slots;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void addDirectSuperslot(Slot slot, Slot superslot) {
        try {
            delegate.addDirectSuperslot(slot, superslot, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void removeDirectSuperslot(Slot slot, Slot superslot) {
        try {
            delegate.removeDirectSuperslot(slot, superslot, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void moveDirectSubslot(Slot slot, Slot subslot, int index) {
        try {
            delegate.moveDirectSubslot(slot, subslot, index, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectTypes(Instance instance) {
        try {
            List clses = delegate.getDirectTypes(instance, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getTypes(Instance instance) {
        try {
            Set clses = delegate.getTypes(instance, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getDirectInstances(Cls cls) {
        try {
            List instances = delegate.getDirectInstances(cls, session);
            localize(instances);
            return instances;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getInstances(Cls cls) {
        try {
            Set instances = delegate.getInstances(cls, session);
            localize(instances);
            return instances;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void addDirectType(Instance instance, Cls type) {
        try {
            delegate.addDirectType(instance, type, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void removeDirectType(Instance instance, Cls type) {
        try {
            delegate.removeDirectType(instance, type, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public List getEvents() {
        try {
            List events = delegate.getEvents(session);
            localize(events);
            return events;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set executeQuery(Query query) {
        try {
            Set frames = delegate.executeQuery(query, session);
            localize(frames);
            return frames;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getReferences(Object object) {
        try {
            Set references = delegate.getReferences(object, session);
            localize(references);
            return references;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getMatchingReferences(String string, int maxMatches) {
        try {
            Set references = delegate.getMatchingReferences(string, maxMatches, session);
            localize(references);
            return references;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getFramesWithDirectOwnSlotValue(Slot slot, Object value) {
        try {
            Set frames = delegate.getFramesWithDirectOwnSlotValue(slot, value, session);
            localize(frames);
            return frames;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getFramesWithAnyDirectOwnSlotValue(Slot slot) {
        try {
            Set frames = delegate.getFramesWithAnyDirectOwnSlotValue(slot, session);
            localize(frames);
            return frames;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getFramesWithMatchingDirectOwnSlotValue(Slot slot, String value, int maxMatches) {
        try {
            Set frames = delegate.getFramesWithMatchingDirectOwnSlotValue(slot, value, maxMatches, session);
            localize(frames);
            return frames;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getClsesWithDirectTemplateSlotValue(Slot slot, Object value) {
        try {
            Set clses = delegate.getClsesWithDirectTemplateSlotValue(slot, value, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getClsesWithAnyDirectTemplateSlotValue(Slot slot) {
        try {
            Set clses = delegate.getClsesWithAnyDirectTemplateSlotValue(slot, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getClsesWithMatchingDirectTemplateSlotValue(Slot slot, String value, int maxMatches) {
        try {
            Set clses = delegate.getClsesWithMatchingDirectTemplateSlotValue(slot, value, maxMatches, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getClsesWithDirectTemplateFacetValue(Slot slot, Facet facet, Object value) {
        try {
            Set clses = delegate.getClsesWithDirectTemplateFacetValue(slot, facet, value, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getClsesWithMatchingDirectTemplateFacetValue(Slot slot, Facet facet, String value, int maxMatches) {
        try {
            Set clses = delegate.getClsesWithMatchingDirectTemplateFacetValue(slot, facet, value, maxMatches, session);
            localize(clses);
            return clses;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public Set getDirectOwnSlotValuesClosure(Frame frame, Slot slot) {
        try {
            Set values = delegate.getDirectOwnSlotValuesClosure(frame, slot, session);
            localize(values);
            return values;
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public boolean beginTransaction(String name) {
        try {
            return delegate.beginTransaction(name, session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public boolean commitTransaction() {
        try {
            return delegate.commitTransaction(session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public boolean rollbackTransaction() {
        try {
            return delegate.rollbackTransaction(session);
        } catch (RemoteException e) {
            throw convertException(e);
        }
    }

    public void close() {
        delegate = null;
    }
}
